/*****************************************************************************/
/* Kernel Hooks Interface.                                    		     */
/* Version 1.3                                                               */
/* Author: Vamsi Krishna S. vamsi_krishna@in.ibm.com 				     */
/*                                                                           */
/* A sample kernel module to register a callk hook handler.		     */
/*                                                                           */
/* Refer to Documentation/hook/HOWTO for details.                            */
/*****************************************************************************/
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/hook.h>
#include <linux/dprobes.h>
#include <linux/dprobes_hooks.h>

int exit1(hook_rec_t *, struct dprobes_struct *);
int exit2(hook_rec_t *, struct dprobes_struct *);
int exit3(hook_rec_t *, struct dprobes_struct *);

extern int testhook(void);

hook_rec_t hook1;
hook_rec_t hook2;
hook_rec_t hook3;

int init_module(void)
{
	int rc;

	hook1.hook_exit = &exit1;
	hook1.hook_exit_name = "exit1";
	rc = hook_exit_register(DPROBES_CALLK, &hook1, 1);
	DPRINTK("hook_exit_register exit1 for hook1 returned %u\n", rc);
	if(rc) 
		return rc;

	hook2.hook_exit = &exit2;
	hook2.hook_exit_name = "exit2";
	rc = hook_exit_register(DPROBES_CALLK, &hook2, 1);
	DPRINTK("hook_exit_register exit2 for hook2 returned %u\n", rc);
	if(rc) {
		hook_exit_deregister(&hook1);
		return rc;
	}

	hook3.hook_exit = &exit3;
	hook3.hook_exit_name = "exit3";
	rc = hook_exit_register(DPROBES_CALLK, &hook3, 1);
	DPRINTK("hook_exit_register exit3 for hook1 returned %u\n", rc);
	if(rc) {
		hook_exit_deregister(&hook1);
		hook_exit_deregister(&hook2);
		return rc;
	}

	/* now arm all exits in a single operation */
	hook1.hook_link = &hook2;
	hook2.hook_link = &hook3;
	hook3.hook_link = NULL;

	rc = hook_exit_arm(&hook1);
	DPRINTK("hook_exit_arm returned %u\n", rc);
	if(rc) {
		hook_exit_deregister(&hook1);
		hook_exit_deregister(&hook2);
		hook_exit_deregister(&hook3);
		return rc;
	}
	printk("dprobes: callk handlers registered.\n");

	return 0;
}

void cleanup_module(void)
{
	int rc;

	rc = hook_exit_disarm(&hook1);
	DPRINTK("hook_exit_disarm returned %u\n", rc);

	DPRINTK("deregistering hook exits\n");

	rc = hook_exit_deregister(&hook1);
	DPRINTK("hook_exit_deregister hook1 returned %u\n", rc);


	printk("dprobes: callk handlers deregistered.\n");
	return;
}

/* exit1 exit to test_hook1 */
int exit1(hook_rec_t * hr, struct dprobes_struct * dp)
{
	printk("exit1: hook exit1 entering...\n");
	printk("pgm=%s\n", dp->mod->pgm.name);
	printk("exit1: hook exit1 exit.\n");
	return HOOK_CONTINUE;
}

/* exit2 exit to test_hook2 */
int exit2(hook_rec_t * hr, struct dprobes_struct * dp)
{
	printk("exit2: hook exit2 entered\n");
	printk("exit2: hook exit2 exiting\n");
	return HOOK_CONTINUE;
}

/* exit3 exit to test_hook1 */
int exit3(hook_rec_t * hr, struct dprobes_struct * dp)
{
	printk("exit3: hook exit3 entered, allow testhook to continue\n");
	printk("exit3: hook exit3 exiting\n");
	return HOOK_RETURN;
}
