%{
/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 */

#include <stdio.h>
#include <string.h>
#include <linux/dprobes.h>
#include <linux/dprobes_in.h> 
#include "cmdl.tab.h"

extern char * strip_quotes(char * str);
extern int cmdl_yyinput(char * buf, int max_size);
#undef YY_INPUT
#define YY_INPUT(b, r, ms) (r = cmdl_yyinput(b, ms))
%}

%option noyywrap

%%

-i |
--insert	{ return INSERT; }
--group		{ return GROUPDEF; }
--type		{ return TYPEDEF; }
-s |
--sym		{ return SYM; }
--merge		{ cmdllval.ul = DP_MERGE; return MERGE; }
--replace	{ cmdllval.ul = DP_REPLACE; return REPLACE; }
--stop-cpus	{ return STOPCPUS; }

--dont-verify-opcodes	{ return DONT_VERIFY_OPCODES; }
--autostacktrace	{ return AUTOSTACKTRACE; }

--log-target	{ return LOGTARGET; }
klog		{ return KLOG; }
ltt		{ return LTT; }
com1		{ return COM1; }
com2		{ return COM2; }
evl		{ return EVL; }

-l |
--log		{ return LOG; }
procname	{ return PROCNAME; }
pid		{ return PID; }
uid		{ return UID; }
ssesp		{ return SSESP; }
cseip		{ return CSEIP; }
tsc		{ return TSC; }
all		{ return LOGALL; }


-r |
--remove	{ return REMOVE; }

-b |
--build-ppdf	{ return BUILDPPDF; }

-p |
--apply-ppdf	{ return APPLYPPDF; } 

-o |
--outfile	{ return OUTFILE; }

-q |
--query		{ return QUERY; }
-x |
--extended	{ return EXTENDED; }

-g |
--getvars	{ return GETVARS; }
--index		{ return INDEX; }
--local		{ return LOCAL; }
--global 	{ return GLOBAL; }
--reset		{ return RESET; }
--zero		{ return RESET; }

-h |
--help		{ return HELP; }

-v |
--version	{ return VERSION; }


-a |
--all 		{ return ALL; }

-n |
--name		{ return NAME; }

[0-9]+ 	|
0x[0-9a-f]+	{ cmdllval.ul = strtoul(cmdltext, NULL, 0); return NUMBER; }

\"[^"\n]+["\n] { cmdllval.string = strip_quotes(strdup(cmdltext));
		   return PATH;}

[+a-zA-Z_\/\.][+a-zA-Z0-9_\/\.\-]* { cmdllval.string = strdup(cmdltext);
				   return STRING;}

[ \t]+		{ ;}
-D		{ return DEFINE; }
"="		{ return '='; }
.		{ return cmdltext[0]; }

%%

