%{
/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <asm/unistd.h>
#include <linux/dprobes.h>
#include <linux/dprobes_in.h>
#include "../main.h"
#include "../ppdf.h"

%}

%union {
 	char * string;
 	unsigned short us;
 	unsigned long ul;
}

%token INSERT REMOVE QUERY GETVARS HELP VERSION ALL NAME DEFINE
%token SYM EXTENDED INDEX LOCAL GLOBAL RESET
%token LOG PROCNAME PID UID SSESP CSEIP TSC LOGALL STOPCPUS
%token LOGTARGET KLOG LTT COM1 COM2 EVL
%token APPLYPPDF BUILDPPDF OUTFILE GROUPDEF TYPEDEF
%token DONT_VERIFY_OPCODES
%token AUTOSTACKTRACE

%token <ul> NUMBER
%token <string> STRING PATH
%token <ul> MERGE REPLACE

%type <string> filename name_cmd

%start cmd

%%

cmd:			{ no_options(); }
	| insert_cmd	{ set_cmd(DP_INSERT); }
	| remove_cmd    { set_cmd(DP_REMOVE); }
	| query_cmd	{ set_cmd(DP_QUERY); }
	| getvars_cmd	{ set_cmd(DP_GETVARS); }
	| build_cmd	{ set_cmd(DP_BUILDPPDF); }
	| apply_cmd	{ set_cmd(DP_APPLYPPDF); }
	| HELP		{ set_cmd(DP_HELP); }
	| VERSION	{ set_cmd(DP_VERSION); }
	;

insert_cmd: insert_file		 	
	  | insert_file insert_sub_cmds	
	  ;

insert_file: INSERT filename  	{ set_string_val(&rpnfile, $2); }
	   | INSERT '=' filename{ set_string_val(&rpnfile, $3); }
	   ;

filename: STRING { $$ = $1; }
	| PATH	{ $$ = $1; }
	;

insert_sub_cmds:  insert_sub_cmd
		| insert_sub_cmds insert_sub_cmd
		;
		
insert_sub_cmd:   sym_cmd	
		| merge_replace 
		| log_cmd
		| target_cmd
		| STOPCPUS
		| substitution
		| group_cmd
		| type_cmd
		| AUTOSTACKTRACE { set_cmd(DP_AUTOSTACKTRACE); }
		| DONT_VERIFY_OPCODES { set_cmd(DP_DONT_VERIFY_OPCODES); }
		;

sym_cmd:  SYM filename		{ set_string_val(&symfile, $2); }
	| SYM '=' filename	{ set_string_val(&symfile, $3); }
	;

merge_replace:	  MERGE	  { set_merge_replace($1); }
		| REPLACE { set_merge_replace($1); }
		;

log_cmd:  LOG log_options
	| LOG '=' log_options
	;

log_options: log_option
	   | log_option log_options
	   ;

log_option: PROCNAME	{ set_cmd(DP_LOG_PROCNAME); }
	  | PID		{ set_cmd(DP_LOG_PID); }
	  | UID		{ set_cmd(DP_LOG_UID); }
	  | SSESP	{ set_cmd(DP_LOG_SS_ESP); }
	  | CSEIP	{ set_cmd(DP_LOG_CS_EIP); }
	  | TSC		{ set_cmd(DP_LOG_TSC); }
	  | LOGALL	{ set_cmd(DP_LOG_ALL); }
	  ;

target_cmd: LOGTARGET target_option
	  | LOGTARGET '=' target_option
	  ;

target_option: KLOG	{ set_cmd(DP_LOG_TARGET_KLOG); }
	     | LTT	{ set_cmd(DP_LOG_TARGET_LTT); }
	     | COM1	{ set_cmd(DP_LOG_TARGET_COM1); }
	     | COM2	{ set_cmd(DP_LOG_TARGET_COM2); }
	     | EVL	{ set_cmd(DP_LOG_TARGET_EVL); }
	     ;

substitution: DEFINE STRING '=' STRING { add_define($2, $4); }
	    | DEFINE STRING '=' NUMBER { add_define_num($2, $4); }
	    ;

remove_cmd: REMOVE what
	  ;

what:     ALL		{ set_cmd(DP_ALL); }
	| name_cmd	{ set_string_val(&modname, $1); }
	;

name_cmd: NAME filename		{ $$ = $2; }
	| NAME '=' filename	{ $$ = $3; }
	;

query_cmd: QUERY what
	 | QUERY what EXTENDED	{ set_cmd(DP_QUERY_EXTENDED); }
         | QUERY EXTENDED what  { set_cmd(DP_QUERY_EXTENDED); }
         ;

getvars_cmd: GETVARS what
           | GETVARS what getvars_sub_cmds
           | GETVARS getvars_sub_cmds what
           ;

getvars_sub_cmds: getvars_sub_cmd
		| getvars_sub_cmd getvars_sub_cmds
		;

getvars_sub_cmd:  LOCAL		{ set_cmd(DP_GETVARS_LOCAL); }
		| GLOBAL	{ set_cmd(DP_GETVARS_GLOBAL); }
		| RESET		{ set_cmd(DP_GETVARS_RESET); }
		| index_cmd	{ set_cmd(DP_GETVARS_INDEX); }
		;

index_cmd: INDEX NUMBER			{ set_range($2, $2); }
	 | INDEX '=' NUMBER		{ set_range($3, $3); }
	 | INDEX NUMBER '-' NUMBER	{ set_range($2, $4); }
	 | INDEX '=' NUMBER '-' NUMBER	{ set_range($3, $5); }
	 ;

build_cmd: build_file
	 | build_file build_sub_cmds
	 ;

build_file: BUILDPPDF filename		{ set_string_val(&rpnfile, $2);}
	  | BUILDPPDF '=' filename	{ set_string_val(&rpnfile, $3);}
	  ;

build_sub_cmds: build_sub_cmd
	      | build_sub_cmd build_sub_cmds
	      ;

build_sub_cmd: build_outfile
	     | sym_cmd
	     | substitution
	     ;

build_outfile:    OUTFILE filename  	{ set_string_val(&ppdffile, $2); }
		| OUTFILE '=' filename	{ set_string_val(&ppdffile, $3); }
		;

apply_cmd: apply_file
	 | apply_file apply_sub_cmds
	 ;

apply_file: APPLYPPDF filename		{ set_string_val(&ppdffile, $2); }
	  | APPLYPPDF '=' filename 	{ set_string_val(&ppdffile, $3); }
	  ;

apply_sub_cmds: apply_sub_cmd
	      | apply_sub_cmd apply_sub_cmds
	      ;

apply_sub_cmd: group_cmd
	     | type_cmd
	     | merge_replace
	     | log_cmd
	     | target_cmd
	     | STOPCPUS
	     ;

group_cmd: GROUPDEF groupname
	| GROUPDEF '=' groupname
	;

type_cmd: TYPEDEF typename
	| TYPEDEF '=' typename
	;

groupname: STRING		{ add_rpn_group($1); }
	 | groupname STRING 	{ add_rpn_group($2); }
	 | groupname ',' STRING { add_rpn_group($3); }
	 ;

typename: STRING		{ add_rpn_type($1); }
	| typename STRING 	{ add_rpn_type($2); }
	| typename ',' STRING 	{ add_rpn_type($3); }
	;

%%

int cmdl_yyinput(char * buf, int max_size)
{
	int remaining = cmdline_end - cmdline_ptr;
	if (max_size > remaining)
		max_size = remaining;

	if (max_size > 0) {
		memcpy(buf, cmdline_ptr, max_size);
		cmdline_ptr += max_size;
	}
	return max_size;
}

/* some of the cmdline options are arch-specific like SSESP etc. */
extern void display_version(void);
void display_help(void)
{
display_version();
printf("\n");
printf("dprobes {--help} \n");
printf("\n");
printf("dprobes {--version} \n");
printf("\n");
printf("dprobes {--insert -i <probe-program-file>}\n"
	"\t[-D key = value]\n"
	"\t[--sym -s <symbol-file>]\n"
	"\t[--group <group_list>]\n"
	"\t[--type <type_list>]\n"
	"\t[--merge | --replace]\n"
	"\t[--log -l <[PROCNAME [UID [SSESP [CSEIP [TSC]]]]] | ALL> ]\n"
	"\t[--log-target <KLOG | LTT | COM1 | COM2 | EVL>]\n");
printf("\n");
printf("dprobes {--remove -r} { --name -n <loadable_module_name> | --all -a}\n");
printf("\n");
printf("dprobes {--getvars -g} { --name -n <loadable_module_name> | --all -a}\n"
	"\t[--index=index1-index2]\n"
	"\t[--local] [--global] [--reset]\n");
printf("\n");
printf("dprobes {--query -q} { --name -n <loadable_module_name> | --all -a}\n"
	"\t[--extended -x] \n");
printf("\n");
printf("dprobes {--build-ppdf -b <probe-program-file> }\n"
	"\t[--outfile -o <out-file>]\n"
	"\t[-D key = value]\n"
	"\t[--sym -s <symbol-file>]\n");
printf("\n");
printf("dprobes {--apply-ppdf -p <ppdf-file> }\n"
	"\t[--group <group_list> ]\n"
	"\t[--type <type_list> ]\n"
	"\t[--merge | --replace]\n"
	"\t[--log -l <[PROCNAME [UID [SSESP [CSEIP [TSC]]]]] | ALL> ]\n"
	"\t[--log-target <KLOG | LTT | COM1 | COM2 | EVL>]\n");
}

