%{
/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>

#include <linux/dprobes.h>
#include <linux/dprobes_in.h>

#include "../main.h"

%}

%union {
 	char * string;
	uint8_t u8;
 	uint16_t u16;
 	unsigned long ul;
 	struct regop_struct regop;
}

%token NAME VARS GVARS MAJOR LOGMAX JMPMAX TYPEDEF GROUPDEF ID
%token MODTYPE KERNEL KMOD USER
%token LOGONFAULT YES NO
%token MINOR OFFSET ADDRESS OPCODE PASSCOUNT MAXHITS TYPE GROUP
%token WATCHPOINT X RW W IO
%token NOP
%token JMP JLT JLE JGT JGE JZ JNZ LOOP CALL RET PROC ENDPROC CALLK
%token ABORT REMOVE SUSPEND RESUME EXIT
%token SETMAJ SETMIN
%token LOG STR ARS MRF MRS B W D
%token ADD MUL AND OR XOR NEG SUB XCHG ROL ROR SHL SHR DUP DIV IDIV PBL PBR PZL PZR
%token VFYR VFYRW
%token PUSH POP ROS
%token BIF WIF DIF QIF IOB IOW IOD PID PROCID TASK R U
%token MEM U8 U16 U32 U64
%token SX UX RX EX_MASK LP PLP EX_LOGMAX ST PURGE TRACE AUTOSTACKTRACE PV
%token SBP TSP COPY STP SAVE RESTORE
%token CS DS ES FS GS SS EAX EBX ECX EDX EDI ESI EIP ESP EBP EFLAGS
%token TR LDTR GDTR IDTR CR0 CR1 CR2 CR3 CR4 DR0 DR1 DR2 DR3 DR6 DR7
%token CPUID MSR
%token FR0 FR1 FR2 FR3 FR4 FR5 FR6 FR7 FCW FSW FTW FIP FCS FDP FDS
%token XMM0 XMM1 XMM2 XMM3 XMM4 XMM5 XMM6 XMM7 MXCSR
%token MOVE LV GV INC DEC
%token SEG2LIN

%token <ul> NUMBER
%token <string> STRING PATH 

%type <u8> jmps call no_oprnd u8_oprnd u16_oprnd yes_no 
%type <u16> regr regu regu_modifiable regsys regu_non_modifiable
%type <u16> mod_type watch_type
%type <string> beginproc
%type <string> filename
%type <string> label
%type <string> groups types
%type <regop> reg_ins
%type <ul> symbol number_expr symbol_expr
%left '+' '-'
%left '*' '/'
%nonassoc UMINUS

%start rpn_file

%%

rpn_file: file_header probe_definitions
	;

file_header: file_hdr { end_file_header(); }
	;

file_hdr: file_hdr_entry
	| file_hdr_entry file_hdr
 	;

file_hdr_entry: NAME '=' filename { pgm.name = strdup($3);}
		| VARS '=' NUMBER { set_lvars($3); }
		| GVARS '=' NUMBER { set_gvars($3); }
	 	| MAJOR '=' NUMBER 
			{ 
				if (!$3) rpnerror("Major 0 reserved");
				else pgm.major = $3;
			}
		| ID '=' NUMBER { pgm.id = $3; }
		| LOGMAX '=' NUMBER { pgm.logmax = $3; }
		| EX_LOGMAX '=' NUMBER { pgm.ex_logmax = $3; }
		| JMPMAX '=' NUMBER { pgm.jmpmax = $3; }
		| MODTYPE '=' mod_type { pgm.flags |= $3; dp_mod_type = $3; }
		| LOGONFAULT '=' yes_no { logonfault = $3; }
		| AUTOSTACKTRACE '=' yes_no { pgm.autostacktrace = $3; }
		| groupdefs
		| typedefs
	 	;

filename: STRING { $$ = $1; }
	| PATH	{ $$ = $1; }
	;

mod_type: USER { $$ = DP_MODTYPE_USER; }
	| KERNEL { $$ = DP_MODTYPE_KERNEL; }
	| KMOD { $$ = DP_MODTYPE_KMOD; }
	;

yes_no:   YES { $$ = 1; }
	| NO  { $$ = 0; }
	;

groupdefs: GROUPDEF '=' groups
	  ;

groups:  STRING { add_group($1); }
	| STRING groups { add_group($1); }
	;

typedefs: TYPEDEF '=' types
	  ;

types:  STRING { add_type($1); }
	| STRING types { add_type($1); }
	;

probe_definitions: probe_definition
 		 | probe_definition probe_definitions
		 ;

probe_definition: probe_hdr pointcode
		| probe_hdr pointcode procedures
		;

probe_hdr: first_probe_hdr_entry 
	 | first_probe_hdr_entry other_probe_hdr_entries
	 ;
other_probe_hdr_entries:  probe_hdr_entry
			| probe_hdr_entry other_probe_hdr_entries 
			;

first_probe_hdr_entry:    OFFSET '=' number_expr { start_probe($3, 0, 0); }
			| OFFSET '=' symbol_expr { start_probe($3, 0, 0); }
			| ADDRESS '=' number_expr { start_probe($3, 0, DP_ADDRESS_ABSOLUTE); } 
			| ADDRESS '=' number_expr ':' number_expr
				{ 
					start_probe($3, ($5 - $3), DP_ADDRESS_ABSOLUTE); 
				} 
			| ADDRESS '=' symbol_expr 
				{	 
				 	start_probe($3, 0, DP_ADDRESS_ABSOLUTE); 
					fix_address();
				}
      			| ADDRESS '=' symbol_expr ':' symbol_expr
				{ 
					start_probe($3, ($5 - $3), DP_ADDRESS_ABSOLUTE);
					fix_address();
				}
		     	;

probe_hdr_entry: MINOR '=' NUMBER {point[cur_point].minor = (unsigned short)$3;}
	| OPCODE '=' NUMBER { probe_on_opcode((unsigned char)$3); }
	| PASSCOUNT '=' NUMBER { point[cur_point].passcount = (long) $3; }
	| MAXHITS '=' NUMBER { point[cur_point].maxhits = $3; }
	| GROUP '=' STRING { point[cur_point].group = group_index($3); }
	| TYPE '=' STRING { point[cur_point].type = type_index($3); }
	| LOGONFAULT '=' yes_no { point[cur_point].logonfault = $3; }
	| WATCHPOINT '=' watch_type 
		{
			set_probe_type(); 
			point[cur_point].probe |= $3; 
		}
	| EX_MASK '=' NUMBER { point[cur_point].ex_mask = $3; }
	;

watch_type: X { $$ = DP_WATCHTYPE_EXECUTE; }
	| RW { $$ = DP_WATCHTYPE_RDWR; }
	| W { $$ = DP_WATCHTYPE_WRITE; }
	| IO { $$ = DP_WATCHTYPE_IO; }
	;

symbol_expr: symbol { $$ = $1; probe_by_symbol = 1; }
	 | symbol '+' number_expr { $$ = $1 + $3; }
	 | symbol '-' number_expr { $$ = $1 - $3; }
	 ;

symbol: STRING { $$ = lookup_symbol_value($1); }

number_expr: NUMBER { $$ = $1; }
	| number_expr '+' number_expr { $$ = $1 + $3; }
	| number_expr '-' number_expr { $$ = $1 - $3; }
	| number_expr '*' number_expr { $$ = $1 * $3; }
	| number_expr '/' number_expr 
	 	{
			if ($3 == 0) { 
				rpnerror("Divide by zero");
			}
			$$ = $1 / $3;
		}
	| '-' number_expr %prec UMINUS { $$ = -$2; }
	| '(' number_expr ')' { $$ = $2; }
	;


pointcode: code { set_opcode(DP_EXIT); fixup_jmps(); end_probe(); }
	 ;

procedures: procedure
	  | procedure procedures 
	  ;

procedure: beginproc code ENDPROC { set_opcode(DP_RET); fixup_jmps(); }
	 ;

beginproc: PROC label { add_call_target($2); }
	 ;

code:	  instruction
	| instruction code 
	;

instruction: jmps label
	   	{
	   		set_opcode_oprnd($1, long, 0);
			add_jmp($2);
	   	}
	   | call label
	   	{
	   		set_opcode_oprnd($1, long, 0);
			add_call($2);
	   	}
	   | no_oprnd { set_opcode($1); }
	   | u8_oprnd NUMBER	{ set_opcode_oprnd($1, uint8_t, $2); }
	   | u16_oprnd NUMBER	{ set_opcode_oprnd($1, uint16_t, $2); }
	   | PUSH B ',' NUMBER	{ set_opcode_oprnd(DP_PUSH, unsigned long, $4); }
	   | PUSH W ',' NUMBER	{ set_opcode_oprnd(DP_PUSH, unsigned long, $4); }
	   | PUSH D ',' NUMBER	{ set_opcode_oprnd(DP_PUSH, unsigned long, $4); }
	   | PUSH number_expr 
		{ 
			set_opcode_oprnd(DP_PUSH, unsigned long, $2); 
		}
	   | PUSH symbol_expr 
		{
			if (pgm.flags & DP_MODTYPE_KERNEL) {
				set_opcode_oprnd(DP_PUSH, unsigned long, fix_offset($2)); 
			}
			else
				rpnerror("Symbol expressions are supported for kernel mode probes only"); 
		}
	   | reg_ins { set_opcode_oprnd($1.opcode, uint16_t, $1.index); }
	   | label ':' { add_jmp_target($1); }
	   ;

label: STRING { $$ = $1; }
	;

jmps:	  JMP { $$ = DP_JMP; }
	| JLT { $$ = DP_JLT; }
	| JLE { $$ = DP_JLE; }
	| JGT { $$ = DP_JGT; }
	| JGE { $$ = DP_JGE; }
	| JZ  { $$ = DP_JZ; }
	| JNZ { $$ = DP_JNZ; }
	| LOOP { $$ = DP_LOOP; }
	| SX { $$ = DP_SX; }
	;

call:	  CALL { $$ = DP_CALL; }
	;

no_oprnd: NOP  { $$ = DP_NOP; }
	| SEG2LIN { $$ = DP_SEG2LIN; }
	| ABORT   { $$ = DP_ABORT; }
	| REMOVE  { $$ = DP_REM; }
	| SUSPEND { $$ = DP_SUSPEND; }
	| RESUME  { $$ = DP_RESUME; }
	| EXIT    { $$ = DP_EXIT; }
	| RET	  { $$ = DP_RET; }
	| SETMAJ  { $$ = DP_SETMAJ; }
	| SETMIN  { $$ = DP_SETMIN; }
	| LOG STR { $$ = DP_LOG_STR; }
	| LOG MRF { $$ = DP_LOG_MRF; }
	| VFYR    { $$ = DP_VFY_R; }
	| VFYRW   { $$ = DP_VFY_RW; }
	| ADD  { $$ = DP_ADD; }
	| MUL  { $$ = DP_MUL; }
	| DIV  { $$ = DP_DIV; }
	| IDIV { $$ = DP_IDIV; }
	| AND  { $$ = DP_AND; }
	| OR   { $$ = DP_OR; }
	| XOR  { $$ = DP_XOR; }
	| NEG  { $$ = DP_NEG; }
	| SUB  { $$ = DP_SUB; }
	| PBL  { $$ = DP_PBL; }
	| PBR  { $$ = DP_PBR; }
	| PZL  { $$ = DP_PZL; }
	| PZR  { $$ = DP_PZR; }
	| XCHG { $$ = DP_XCHG; }
	| ROL  { $$ = DP_ROL; }
	| ROR  { $$ = DP_ROR; }
	| SHL  { $$ = DP_SHL; }
	| SHR  { $$ = DP_SHR; }
	| DUP  { $$ = DP_DUP; }
	| PUSH BIF { $$ = DP_PUSH_MEM_U8; }
	| PUSH WIF { $$ = DP_PUSH_MEM_U16; }
	| PUSH DIF { $$ = DP_PUSH_MEM_U32; }
	| PUSH QIF { $$ = DP_PUSH_MEM_U64; }
	| POP BIF  { $$ = DP_POP_MEM_U8; }
	| POP WIF  { $$ = DP_POP_MEM_U16; }
	| POP DIF  { $$ = DP_POP_MEM_U32; }
	| POP QIF  { $$ = DP_POP_MEM_U64; }
	| PUSH MEM ',' U8  { $$ = DP_PUSH_MEM_U8; }
	| PUSH MEM ',' U16 { $$ = DP_PUSH_MEM_U16; }
	| PUSH MEM ',' U32 { $$ = DP_PUSH_MEM_U32; }
	| PUSH MEM ',' U64 { $$ = DP_PUSH_MEM_U64; }
	| POP MEM ',' U8   { $$ = DP_POP_MEM_U8; }
	| POP MEM ',' U16  { $$ = DP_POP_MEM_U16; }
	| POP MEM ',' U32  { $$ = DP_POP_MEM_U32; }
	| POP MEM ',' U64  { $$ = DP_POP_MEM_U64; }
	| PUSH IOB { $$ = DP_PUSH_IO_U8; }
	| PUSH IOW { $$ = DP_PUSH_IO_U16; }
	| PUSH IOD { $$ = DP_PUSH_IO_U32; }
	| POP IOB  { $$ = DP_POP_IO_U8; }
	| POP IOW  { $$ = DP_POP_IO_U16; }
	| POP IOD  { $$ = DP_POP_IO_U32; }
	| PUSH IO ',' U8  { $$ = DP_PUSH_IO_U8; }
	| PUSH IO ',' U16 { $$ = DP_PUSH_IO_U16; }
	| PUSH IO ',' U32 { $$ = DP_PUSH_IO_U32; }
	| POP IO ',' U8   { $$ = DP_POP_IO_U8; }
	| POP IO ',' U16  { $$ = DP_POP_IO_U16; }
	| POP IO ',' U32  { $$ = DP_POP_IO_U32; }
	| PUSH LV  { $$ = DP_PUSH_LV; }
	| POP LV   { $$ = DP_POP_LV; }
	| MOVE LV  { $$ = DP_MOVE_LV; }
	| INC LV   { $$ = DP_INC_LV; }
	| DEC LV   { $$ = DP_DEC_LV; }
	| PUSH GV  { $$ = DP_PUSH_GV; }
	| POP GV   { $$ = DP_POP_GV; }
	| MOVE GV  { $$ = DP_MOVE_GV; }
	| INC GV   { $$ = DP_INC_GV; }
	| DEC GV   { $$ = DP_DEC_GV; }
	| PUSH PROCID  { $$ = DP_PUSH_PROCID; }
	| PUSH PID  { $$ = DP_PUSH_PID; }
	| PUSH TASK { $$ = DP_PUSH_TASK; }
	| CALLK	    { $$ = DP_CALLK; }
	| UX { $$ = DP_UX; }
	| RX { $$ = DP_RX; }
	| PUSH X { $$ = DP_PUSH_X; }
	| PUSH LP { $$ = DP_PUSH_LP; }
	| PUSH PLP { $$ = DP_PUSH_PLP; }
	| POP LP { $$ = DP_POP_LP; }
	| LOG ST { $$ = DP_LOG_ST; }
	| PURGE ST { $$ = DP_PURGE_ST; }
	| TRACE LV { $$ = DP_TRACE_LV; }
	| TRACE GV { $$ = DP_TRACE_GV; }
	| TRACE PV { $$ = DP_TRACE_PV; }
	| PUSH SBP { $$ = DP_PUSH_SBP; }
	| POP SBP { $$ = DP_POP_SBP; }
	| COPY SBP { $$ = DP_COPY_SBP; }
	| PUSH TSP { $$ = DP_PUSH_TSP; }
	| POP TSP { $$ = DP_POP_TSP; }
	| COPY TSP { $$ = DP_COPY_TSP; }
	| PUSH STP { $$ = DP_PUSH_STP; }
	| POP STP { $$ = DP_POP_STP; }
	| SAVE SBP { $$ = DP_SAVE_SBP; }
	| RESTORE SBP { $$ = DP_RESTORE_SBP; }
	| SAVE TSP { $$ = DP_SAVE_TSP; }
	| RESTORE TSP { $$ = DP_RESTORE_TSP; }
	| LOG { $$ = DP_LOG; }
	| LOG LV { $$ = DP_LOG_LV; }
	| LOG GV { $$ = DP_LOG_GV; }
	;

u8_oprnd: ROL  { $$ = DP_ROL_I; }
	| ROR  { $$ = DP_ROR_I; }
	| SHL  { $$ = DP_SHL_I; }
	| SHR  { $$ = DP_SHR_I; }
	| EXIT { $$ = DP_EXIT_N; }
	| DUP  { $$ = DP_DUP_I; }
	| PBL { $$ = DP_PBL_I; }
	| PBR { $$ = DP_PBR_I; }
	| PZL { $$ = DP_PZL_I; }
	| PZR { $$ = DP_PZR_I; }
	| ROS { $$ = DP_ROS; }
	;

u16_oprnd: SETMAJ  { $$ = DP_SETMAJ_I; }
	| SETMIN { $$ = DP_SETMIN_I; }
	| PUSH R ',' { $$ = DP_PUSH_R; }
	| PUSH U ',' { $$ = DP_PUSH_U; }
	| POP  R ',' { $$ = DP_POP_R; }
	| POP  U ',' { $$ = DP_POP_U; }
	| PUSH LV ',' { $$ = DP_PUSH_LVI; }
	| POP  LV ',' { $$ = DP_POP_LVI; }
	| MOVE LV ',' { $$ = DP_MOVE_LVI; }
	| INC LV ','  { $$ = DP_INC_LVI; }
	| DEC LV ','  { $$ = DP_DEC_LVI; }
	| PUSH GV ',' { $$ = DP_PUSH_GVI; }
	| POP  GV ',' { $$ = DP_POP_GVI; }
	| MOVE GV ',' { $$ = DP_MOVE_GVI; }
	| INC GV ','  { $$ = DP_INC_GVI; }
	| DEC GV ','  { $$ = DP_DEC_GVI; }
	| LOG  B ',' { $$ = DP_LOG_I; }
	| LOG  W ',' { $$ = DP_LOG_I; }
	| LOG  D ',' { $$ = DP_LOG_I; }
	| LOG        { $$ = DP_LOG_I; }
	| PUSH SBP ',' { $$ = DP_PUSH_SBP_I; }
	| POP SBP ',' { $$ = DP_POP_SBP_I; }
	| PUSH TSP ',' { $$ = DP_PUSH_TSP_I; }
	| POP TSP ',' { $$ = DP_POP_TSP_I; }
	| COPY SBP ',' { $$ = DP_COPY_SBP_I; }
	| COPY TSP ',' { $$ = DP_COPY_TSP_I; }
	;

reg_ins:  PUSH U ',' regu { $$.opcode = DP_PUSH_U; $$.index = $4; }
	| POP U ',' regu_modifiable { $$.opcode = DP_POP_U; $$.index = $4; }
	| PUSH R ',' regr { $$.opcode = DP_PUSH_R; $$.index = $4; }
	| POP R ',' regu_modifiable { $$.opcode = DP_POP_R; $$.index = $4; }
	;


regr:	  CS  { $$ = DP_CS; }
        | SS  { $$ = DP_SS; }
        | EIP { $$ = DP_EIP; }
        | ESP { $$ = DP_ESP; }
        | EFLAGS { $$ = DP_EFLAGS; }
        | regu_modifiable { $$ = $1; }
	| regsys { $$ = $1; }
	;

regu:     CS  { $$ = DP_CS; }
	| SS  { $$ = DP_SS; }
	| EIP { $$ = DP_EIP; }
	| ESP { $$ = DP_ESP; }
	| EFLAGS { $$ = DP_EFLAGS; }
	| regu_modifiable { $$ = $1; }
	| regu_non_modifiable { $$ = $1; }
	;

regu_modifiable: DS { $$ = DP_DS; }
	| ES { $$ = DP_ES; }
	| FS { $$ = DP_FS; }
	| GS { $$ = DP_GS; }
	| EAX { $$ = DP_EAX; }
	| EBX { $$ = DP_EBX; }
	| ECX { $$ = DP_ECX; }
	| EDX { $$ = DP_EDX; }
	| EDI { $$ = DP_EDI; }
	| ESI { $$ = DP_ESI; }
	| EBP { $$ = DP_EBP; }
	;

regu_non_modifiable: FR0 { $$ = DP_FR0; }
	| FR1 { $$ = DP_FR1; }
	| FR2 { $$ = DP_FR2; }
	| FR3 { $$ = DP_FR3; }
	| FR4 { $$ = DP_FR4; }
	| FR5 { $$ = DP_FR5; }
	| FR6 { $$ = DP_FR6; }
	| FR7 { $$ = DP_FR7; }
	| FCW { $$ = DP_FCW; }
	| FSW { $$ = DP_FSW; }
	| FTW { $$ = DP_FTW; }
	| FIP { $$ = DP_FIP; }
	| FCS { $$ = DP_FCS; }
	| FDP { $$ = DP_FDP; }
	| FDS { $$ = DP_FDS; }
	| XMM0 { $$ = DP_XMM0; }
	| XMM1 { $$ = DP_XMM1; }
	| XMM2 { $$ = DP_XMM2; }
	| XMM3 { $$ = DP_XMM3; }
	| XMM4 { $$ = DP_XMM4; }
	| XMM5 { $$ = DP_XMM5; }
	| XMM6 { $$ = DP_XMM6; }
	| XMM7 { $$ = DP_XMM7; }
	| MXCSR { $$ = DP_MXCSR; }
	;

regsys: TR { $$ = DP_TR; }
	| LDTR { $$ = DP_LDTR; }
	| GDTR { $$ = DP_GDTR; }
	| IDTR { $$ = DP_IDTR; }
	| CR0 { $$ = DP_CR0; }
	| CR1 { $$ = DP_CR1; }
	| CR2 { $$ = DP_CR2; }
	| CR3 { $$ = DP_CR3; }
	| CR4 { $$ = DP_CR4; }
	| DR0 { $$ = DP_DR0; }
	| DR1 { $$ = DP_DR1; }
	| DR2 { $$ = DP_DR2; }
	| DR3 { $$ = DP_DR3; }
	| DR6 { $$ = DP_DR6; }
	| DR7 { $$ = DP_DR7; }
	| CPUID { $$ = DP_CPUID; }
	| MSR { $$ = DP_MSR; }
	;

%%
void rpnerror(const char * err)
{
	printf("%s(%d) %s\n", filename, rpnlineno, err);
	exit(1);
}

#define NUM_INVALID_OPCODES 3
static unsigned char invalid_opcode[] = { 
		0xcc,	/* INT 3 */
		0xcd,	/* INT n */
		0xce	/* INTO */
		};

static int valid_opcode (unsigned char op)
{
	int i = NUM_INVALID_OPCODES;
	for (i = 0; i < NUM_INVALID_OPCODES; i++) {
		if (invalid_opcode[i] == op)
			return 0;
	}
	return 1;
}

int probe_on_opcode(unsigned char op)
{
	char buff[80];
	if (valid_opcode(op)) {
		point[cur_point].opcode = op; 
	} else {
		sprintf(buff, "probe not allowed on opcode 0x%x\n", op);
		rpnerror(buff);
	}
	return 0;
}

