#include <stdint.h>

#define MAX_TYPES	256
#define MAX_GROUPS	256

struct regop_struct {
        uint8_t opcode;
        uint16_t index;
};

typedef struct {
	char * name;
	unsigned short offset;
	int lineno;
} list_entry_t;

typedef struct {
	char *type_name;
	int max;
	int count;
	list_entry_t *entry;
} label_table_t;

extern int init_syms(const char * filename);
extern int cleanup_syms(void);
extern unsigned long lookup_symbol_value(const char * name);
extern unsigned short get_kmod_align(void);
extern unsigned short dp_mod_type;
extern char * strip_quotes(char * str);
extern int fatal_error(char * str);

extern char * cmdline;
extern char * cmdline_ptr;
extern char * cmdline_end;
extern int cmdline_len;

extern char * rpnfile;
extern char * ppdffile;
extern char * symfile;
extern char * modname;

extern void set_cmd(unsigned long opt);
extern void add_rpn_group(char * name);
extern void add_rpn_type(char * name);

#define NUM_POINTS	256
#define CODE_BUF_SIZE	4096

extern unsigned short dp_mod_type;

extern unsigned int rpnlineno;

extern struct dp_pgm_struct pgm;
extern struct dp_point_struct *point;
extern unsigned int cur_point;

extern char * filename;

extern unsigned char logonfault;
extern int probe_by_symbol;

extern unsigned char * rpn_code;
extern unsigned int rpn_ip;

extern int probe_on_opcode(unsigned char); 
extern void start_probe(unsigned long, unsigned long, unsigned long);

extern void no_options(void);
extern void set_string_val(char **dest, char * src);
extern void set_merge_replace(unsigned long opt);
extern void add_define(char * key, char * val);
extern void add_define_num(char * key, unsigned long ulval);
extern void set_range(unsigned long from, unsigned long to);
	
extern void add_group(char * name);
extern void add_type(char * name);
extern void ins_size(int s);
extern void end_file_header(void);
extern void end_probe(void);
extern void set_lvars(unsigned long n);
extern void set_gvars(unsigned long n);
extern unsigned long fix_offset(unsigned long addr);
extern void fix_address(void);
extern void fixup_jmps(void);
extern int group_index(char * name);
extern int type_index(char * name);
extern void set_probe_type(void);
extern void add_call(char * name);
extern void add_jmp(char * name);
extern void add_call_target(char * name);
extern void add_jmp_target(char * name);

	
extern void cmdlerror(const char *);
extern void rpnerror(const char *);

#define set_opcode(op) { \
	ins_size(1); \
	rpn_code[rpn_ip++] = (uint8_t) op; }

#define set_oprnd(type, val) { \
	ins_size(sizeof(type)); \
	*(type *)(&rpn_code[rpn_ip]) = (type)val; \
	rpn_ip += sizeof(type); }

#define set_opcode_nocheck(op) { rpn_code[rpn_ip++] = (uint8_t) op; }

#define set_oprnd_nocheck(type, val) { \
	*(type *)(&rpn_code[rpn_ip]) = (type)val; \
	rpn_ip += sizeof(type); }

#define set_opcode_oprnd(op, type, val) { \
	if ((op >= DP_PUSH_LVI && op <= DP_DEC_LVI) && (val >= pgm.num_lv)) \
		rpnerror("Invalid variable index"); \
	else if((op >= DP_PUSH_GVI && op <= DP_DEC_GVI) && (val >=pgm.num_gv)) \
		rpnerror("Invalid variable index"); \
	ins_size(sizeof(type)+1); \
	set_opcode_nocheck(op); \
	set_oprnd_nocheck(type, val); }

#define NUM_JMPS 	256
#define NUM_CALLS 	256
#define NUM_GROUPS 	256
#define NUM_TYPES 	256

#define MAX_ENTRIES	256

